<html lang="en">

<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta http-equiv="X-UA-Compatible" content="ie=edge">
    <title>{% block title %}{% endblock %}</title>

    <link rel="preconnect" href="https://fonts.googleapis.com">
    <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
    <link href="https://fonts.googleapis.com/css2?family=Open+Sans&display=swap" rel="stylesheet">

    <style>
        body {
            background-color: #eef2f5;
            font-family: "Open Sans", sans-serif;
        }

        div.content {
            min-height: 100%;
            display: flex;
            flex-direction: column;
            align-items: center;
            justify-content: center;
        }

        div.form_wrapper {
            background-color: white;
            flex-grow: 0;
            margin: 0 auto;
            max-width: 25rem;
            width: 90%;
            border: 1px solid rgba(0, 0, 0, 0.1);
            box-sizing: border-box;
            padding: 2rem;
            border-radius: 0.5rem;
        }

        h1 {
            font-size: 1.8rem;
        }

        p.error {
            font-size: 0.9rem;
            color: red;
        }

        label {
            font-size: 0.85rem;
        }

        input,
        label {
            display: block;
            margin: 0.5rem 0;
            width: 100%;
        }

        input,
        button {
            border-radius: 0.2rem;
        }

        input {
            border: 1px solid rgba(0, 0, 0, 0.2);
            padding: 0.5rem;
            margin: 0.5rem 0 0.8rem;
        }

        button {
            background-color: #419EF8;
            border: none;
            color: white;
            cursor: pointer;
            display: block;
            font-size: 0.8rem;
            letter-spacing: 0.05rem;
            padding: 0.5rem;
            text-transform: uppercase;
            width: 100%;
        }
    </style>
</head>
<body>
    <div class="content">
        <div class="form_wrapper">
            {% block content %}{% endblock %}
        </div>
    </div>
</body>
</html>
